@@ -8,6 +8,7 @@ A universal Python parser combinator library inspired by Parsec library of Haske
 __author__ = 'He Tao, sighingnow@gmail.com'
 
 import re
+import warnings
 from functools import wraps
 from collections import namedtuple
 
@@ -42,7 +43,7 @@ class ParseError(RuntimeError):
             return '<out of bounds index {!r}>'.format(self.index)
 
     def __str__(self):
-        return 'expected {} at {}'.format(self.expected, self.loc())
+        return 'expected: {} at {}'.format(self.expected, self.loc())
 
 ##########################################################################
 # Definition the Value model of parsec.py.
@@ -71,6 +72,12 @@ class Value(namedtuple('Value', 'status index value expected')):
             return other
         return Value(True, other.index, self.value + other.value, None)
 
+    def update_index(self, index=None):
+        if index is None:
+            return self
+        else:
+            return Value(self.status, index, self.value, self.expected)
+
     @staticmethod
     def combinate(values):
         '''aggregate multiple values into tuple'''
@@ -111,7 +118,7 @@ class Parser(object):
         return self.fn(text, index)
 
     def parse(self, text):
-        '''Parser a given string `text`.'''
+        '''Parses a given string `text`.'''
         return self.parse_partial(text)[0]
 
     def parse_partial(self, text):
@@ -189,7 +196,7 @@ class Parser(object):
         '''(<<) Ends with a specified parser, and at the end parser consumed the
         end flag.'''
         @Parser
-        def ends_with_parser(text, index):
+        def skip_parser(text, index):
             res = self(text, index)
             if not res.status:
                 return res
@@ -198,7 +205,7 @@ class Parser(object):
                 return Value.success(end.index, res.value)
             else:
                 return Value.failure(end.index, 'ends with {}'.format(end.expected))
-        return ends_with_parser
+        return skip_parser
 
     def ends_with(self, other):
         '''(<) Ends with a specified parser, and at the end parser hasn't consumed
@@ -215,6 +222,20 @@ class Parser(object):
                 return Value.failure(end.index, 'ends with {}'.format(end.expected))
         return ends_with_parser
 
+    def excepts(self, other):
+        '''Fail though matched when the consecutive parser `other` success for the rest text.'''
+        @Parser
+        def excepts_parser(text, index):
+            res = self(text, index)
+            if not res.status:
+                return res
+            lookahead = other(text, res.index)
+            if lookahead.status:
+                return Value.failure(res.index, 'should not be "{}"'.format(lookahead.value))
+            else:
+                return res
+        return excepts_parser
+
     def parsecmap(self, fn):
         '''Returns a parser that transforms the produced value of parser with `fn`.'''
         return self.bind(lambda res: Parser(lambda _, index: Value.success(index, fn(res))))
@@ -262,8 +283,19 @@ class Parser(object):
         '''Implements the `(>>)` operator, means `compose`.'''
         return self.compose(other)
 
+    def __rshift__(self, other):
+        '''Implements the `(>)` operator, means `compose`.'''
+        return self.compose(other)
+
     def __irshift__(self, other):
         '''Implements the `(>>=)` operator, means `bind`.'''
+        warnings.warn('Call to deprecated operator (`>>=`) as it is an in-place '
+                      'operator and offten causing misleading, using (`>=`) for '
+                      'bind() instead.', category=DeprecationWarning)
+        return self.bind(other)
+
+    def __ge__(self, other):
+        '''Implements the `(>=)` operator, means `bind`.'''
         return self.bind(other)
 
     def __lshift__(self, other):
@@ -274,19 +306,22 @@ class Parser(object):
         '''Implements the `(<)` operator, means `ends_with`.'''
         return self.ends_with(other)
 
+    def __truediv__(self, other):
+        return self.excepts(other)
 
-def parse(p, text, index):
+
+def parse(p, text, index=0):
     '''Parse a string and return the result or raise a ParseError.'''
-    return p.parse(text, index)
+    return p.parse(text[index:])
 
 
 def bind(p, fn):
-    '''Bind two parsers, implements the operator of `(>>=)`.'''
+    '''Bind two parsers, implements the operator of `(>=)`.'''
     return p.bind(fn)
 
 
 def compose(pa, pb):
-    '''Compose two parsers, implements the operator of `(>>)`.'''
+    '''Compose two parsers, implements the operator of `(>>)`, or `(>)`.'''
     return pa.compose(pb)
 
 
@@ -329,6 +364,11 @@ def ends_with(pa, pb):
     return pa.ends_with(pb)
 
 
+def excepts(pa, pb):
+    '''Fail `pa` though matched when the consecutive parser `pb` success for the rest text.'''
+    return pa.excepts(pb)
+
+
 def parsecmap(p, fn):
     '''Returns a parser that transforms the produced value of parser with `fn`.'''
     return p.parsecmap(fn)
@@ -420,6 +460,9 @@ def times(p, mint, maxt=None):
         while cnt < maxt:
             res = p(text, index)
             if res.status:
+                if maxt == float('inf') and res.index == index:
+                    # prevent infinite loop, see GH-43
+                    break
                 values = values.aggregate(
                     Value.success(res.index, [res.value]))
                 index, cnt = res.index, cnt + 1
@@ -466,7 +509,7 @@ def optional(p, default_value=None):
             return Value.success(res.index, res.value)
         else:
             # Return the maybe existing default value without doing anything.
-            return Value.success(res.index, default_value)
+            return Value.success(index, default_value)
     return optional_parser
 
 
@@ -486,7 +529,7 @@ def separated(p, sep, mint, maxt=None, end=None):
     '''Repeat a parser `p` separated by `s` between `mint` and `maxt` times.
     When `end` is None, a trailing separator is optional.
     When `end` is True, a trailing separator is required.
-    When `end` is False, a trailing separator is not allowed.
+    When `end` is False, a trailing separator will not be parsed.
     MATCHES AS MUCH AS POSSIBLE.
     Return list of values returned by `p`.'''
     maxt = maxt if maxt else mint
@@ -494,37 +537,37 @@ def separated(p, sep, mint, maxt=None, end=None):
     @Parser
     def sep_parser(text, index):
         cnt, values, res = 0, Value.success(index, []), None
+        prev_values = values
         while cnt < maxt:
-            if end in [False, None] and cnt > 0:
-                res = sep(text, index)
-                if res.status:  # `sep` found, consume it (advance index)
-                    index, values = res.index, Value.success(
-                        res.index, values.value)
-                elif cnt < mint:
-                    return res  # error: need more elemnts, but no `sep` found.
-                else:
-                    break
-
             res = p(text, index)
             if res.status:
-                values = values.aggregate(
+                current_values = values.aggregate(
                     Value.success(res.index, [res.value]))
                 index, cnt = res.index, cnt + 1
-            elif cnt >= mint:
-                break
             else:
-                return res  # error: need more elements, but no `p` found.
-
-            if end is True:
-                res = sep(text, index)
-                if res.status:
-                    index, values = res.index, Value.success(
-                        res.index, values.value)
+                if cnt < mint:
+                    return res  # error: need more elements, but no `p` found.
                 else:
-                    return res  # error: trailing `sep` not found
+                    return values
 
-            if cnt >= maxt:
-                break
+            # consume the sep
+            res = sep(text, index)
+            if res.status:  # `sep` found, consume it (advance index)
+                index = res.index
+                if end in [True, None]:
+                    current_values = current_values.update_index(res.index)
+            else:
+                if cnt < mint or (cnt == mint and end is True):
+                    return res  # error: need more elements, but no `sep` found.
+                else:
+                    if end is True:
+                        # step back
+                        return values
+                    else:
+                        return current_values
+
+            # record the prev values
+            prev_values, values = values, current_values
         return values
     return sep_parser
 
@@ -571,8 +614,19 @@ def sepEndBy1(p, sep):
 ##########################################################################
 
 
+def any():
+    '''Parses a arbitrary character.'''
+    @Parser
+    def any_parser(text, index=0):
+        if index < len(text):
+            return Value.success(index + 1, text[index])
+        else:
+            return Value.failure(index, 'a random char')
+    return any_parser
+
+
 def one_of(s):
-    '''Parser a char from specified string.'''
+    '''Parses a char from specified string.'''
     @Parser
     def one_of_parser(text, index=0):
         if index < len(text) and text[index] in s:
@@ -583,7 +637,7 @@ def one_of(s):
 
 
 def none_of(s):
-    '''Parser a char NOT from specified string.'''
+    '''Parses a char NOT from specified string.'''
     @Parser
     def none_of_parser(text, index=0):
         if index < len(text) and text[index] not in s:
@@ -594,7 +648,7 @@ def none_of(s):
 
 
 def space():
-    '''Parser a whitespace character.'''
+    '''Parses a whitespace character.'''
     @Parser
     def space_parser(text, index=0):
         if index < len(text) and text[index].isspace():
@@ -605,7 +659,7 @@ def space():
 
 
 def spaces():
-    '''Parser zero or more whitespace characters.'''
+    '''Parses zero or more whitespace characters.'''
     return many(space())
 
 
@@ -621,7 +675,7 @@ def letter():
 
 
 def digit():
-    '''Parse a digit character.'''
+    '''Parse a digit.'''
     @Parser
     def digit_parser(text, index=0):
         if index < len(text) and text[index].isdigit():
@@ -632,7 +686,7 @@ def digit():
 
 
 def eof():
-    '''Parser EOF flag of a string.'''
+    '''Parses EOF flag of a string.'''
     @Parser
     def eof_parser(text, index=0):
         if index >= len(text):
@@ -643,7 +697,7 @@ def eof():
 
 
 def string(s):
-    '''Parser a string.'''
+    '''Parses a string.'''
     @Parser
     def string_parser(text, index=0):
         slen, tlen = len(s), len(text)
@@ -658,7 +712,7 @@ def string(s):
 
 
 def regex(exp, flags=0):
-    '''Parser according to a regular expression.'''
+    '''Parses according to a regular expression.'''
     if isinstance(exp, str):
         exp = re.compile(exp, flags)
 
@@ -670,3 +724,59 @@ def regex(exp, flags=0):
         else:
             return Value.failure(index, exp.pattern)
     return regex_parser
+
+
+##########################################################################
+# Useful utility parsers
+##########################################################################
+
+
+def fail_with(message):
+    return Parser(lambda _, index: Value.failure(index, message))
+
+
+def exclude(p: Parser, exclude: Parser):
+    '''Fails parser p if parser `exclude` matches'''
+    @Parser
+    def exclude_parser(text, index):
+        res = exclude(text, index)
+        if res.status:
+            return Value.failure(index, 'something other than %s' %res.value)
+        else:
+            return p(text, index)
+    return exclude_parser
+
+
+def lookahead(p: Parser):
+    '''Parses without consuming'''
+    @Parser
+    def lookahead_parser(text, index):
+        res = p(text, index)
+        if res.status:
+            return Value.success(index, res.value)
+        else:
+            return Value.failure(index, res.expected)
+    return lookahead_parser
+
+
+def unit(p: Parser):
+    '''Converts a parser into a single unit. Only consumes input if the parser succeeds'''
+    @Parser
+    def unit_parser(text, index):
+        res = p(text, index)
+        if res.status:
+            return Value.success(res.index, res.value)
+        else:
+            return Value.failure(index, res.expected)
+    return unit_parser
+
+
+
+def fix(fn):
+    '''Allow recursive parser using the Y combinator trick.
+
+       Note that this version still yields the stack overflow problem, and will be fixed in later version.
+
+       See also: https://github.com/sighingnow/parsec.py/issues/39.
+    '''
+    return (lambda x: x(x))(lambda y: fn(lambda *args: y(y)(*args)))
